/* 
FFT and IFFT of Low-Passed PWM
Based on work of Sam Bobb, Daniel Cornew, Ryan Deeter
Modified by Andrew Dai
*/


/** INCLUDES ***************************************************/
#include "HardwareProfile.h"
#include <plib.h>
#include "dsplib_dsp.h"
#include "fftc.h"
#include "math.h"


/** Constants **************************************************/
#define TRUE 				1
#define FALSE				0

#define PI					(3.14159265)

#define LOOP_TIME_PIN		LATAbits.LATA14	// Timing Pin

#define DESIRED_BAUDRATE    (19200)     	// The desired BaudRate 
	
#define fftc 				fft32c256		//# of samples, from fftc.h, for N = 256 use fft16c256, for N = 1024 use fft16c1024	
#define N 					256				// Also change the log2N variable below!!

// To modify the sampling frequency
#define SAMPLE_FREQ 		1280    		// Also change the timer interupt setup in initInterruptController!

#define PTS_PER_DUTY		(1024)			// set how many times per cycle the duty cycle is changed


/** Function Declarations **************************************/
void initInterruptController();
void initUART2(int pbClk);
void sendDataRS232(void);
void computeFFT(void);
void computeIFFT(void);
void initPWM(void);
void initPWMInterrupt(void);
void initADC(void);
void initSinPWM(void);


/** Global Variables *******************************************/
int computeFFTflag = FALSE; // used to indicate that an fft computation has been requested over rs232
int sampleIndex = 0; 		// keeps track of where we're putting our ADC reading

// this has to be changed for different values of N
int log2N = 8; 				// log2(256) = 8

int32c sampleBuffer[N];		// initialize buffer to collect samples
int32c calcBuffer[N]; 		// initialize buffer to hold old samples
int32c scratch[N];
int32c dout[N]; 			// holds computed FFT until transmission
int32c FFTconjugate[N];		// conjugate of computed FFT

int32 singleSidedFFT[N];
int32 freqVector[N];
int32 FFTPhase[N];

// sine wave related vars
float 	freq = 30; 						// sine wave freq, in Hz, to be generated by PWM
int 	mVpp = 200;						// mV peak-to-peak 
float 	phase = 0;						// phase angle shift
float 	freq2 = 40;						// sine wave freq, in Hz, to be generated by PWM
int 	mVpp2 = 200;					// mV peak-to-peak 
float 	phase2 = 0;						// phase angle shift
int16	duty_cycles[PTS_PER_DUTY] = {0};// vector to hold duty cycles
int 	pwmIndex = 0; 					// increments to PTS_PER_DUTY


/** Main Function **********************************************/
int main(void)
{
	int	pbClk;
	int i;

	// Configure the proper PB frequency and the number of wait states
	pbClk = SYSTEMConfigPerformance(SYS_FREQ);
	
	TRISAbits.TRISA14 = 0;  // Pin for timing
		
	// Allow vector interrupts
	INTEnableSystemMultiVectoredInt();
	
	// Initialize
	mInitAllLEDs();
	initPWM();
	initPWMInterrupt();
	initInterruptController();
	initUART2(pbClk);
	initADC();
	initSinPWM();

	// zero the freqVector, singleSidedFFT, and imaginary part of sample buffer
	for (i=0; i<N; i++)
	{
		freqVector[i] = 0;
		singleSidedFFT[i] = 0;
		FFTPhase[i] = 0;
		sampleBuffer[sampleIndex].im = 0;
	}

	// generate frequency vector
	// this is the x-axis of your single sided fft
	for (i=0; i<N/2; i++)
	{
		freqVector[i] = i*(SAMPLE_FREQ/2)/((N/2) - 1);
	}

	// wait around until you get an fft command from rs232
	// then call computeFFT to generate an fft of the last block of samples
	// then send out rs232.
	// If you want to use FFT without rs232, you can call computeFFT when ever you want in your program.
	while(1)
	{
		
		// For RS232 MATLAB TESTING
		if (computeFFTflag == TRUE)
		{
			computeFFT();
			computeIFFT();
			sendDataRS232();
		}
					
		/*
		// For speed testing
		if (sampleIndex == 0)
		{
			sampleIndex++;
			computeFFT();
			sendDataRS232();
		}
		*/
	}
	
	CloseOC1();


} //end main


/** Interrupt Handlers *****************************************/
// PWM interupt code for the timer 4
// Priority level 6
void __ISR( _TIMER_4_VECTOR, ipl6) T4Interrupt( void)
{
	// set duty cycle 
	SetDCOC1PWM(duty_cycles[pwmIndex]);
	if (pwmIndex == (PTS_PER_DUTY-1)) {
		pwmIndex=0;
	} else {
		pwmIndex++;
	}
	// clear interrupt flag and exit
	mT4ClearIntFlag();
} // end T4 Interrupt

// interrput code for the timer 3
// Priority Level 7
void __ISR( _TIMER_3_VECTOR, ipl7) T3Interrupt( void)
{
	int i;
	//LOOP_TIME_PIN = TRUE;
	// When we used the loop time pin to measure the length of this ISR,
	// we measured 400 ns, so you could sample at over 1 MHz.
	sampleBuffer[sampleIndex].re = ReadADC10(0); // read the ADC into the real part of the samplebuffer
	// increment the sampleIndex
	if (sampleIndex == (N-1))
	{
		sampleIndex = 0;
	}
	else
	{
		sampleIndex++;
	}	 
	//LOOP_TIME_PIN = FALSE;
	// clear interrupt flag and exit
	mT3ClearIntFlag();
} // T3 Interrupt


// UART 2 interrupt handler
// it is set at priority level 2
void __ISR(_UART2_VECTOR, ipl2) IntUart2Handler(void)
{
	char data;
	// Is this an RX interrupt?
	if(mU2RXGetIntFlag())
	{
		// Clear the RX interrupt Flag
		mU2RXClearIntFlag();
		data = ReadUART2();
		// Echo what we just received.
		putcUART2(data);
		switch(data)
		{
			case 'p': // compute and output the FFT
				computeFFTflag = TRUE;
				break;	
		}			
 		// Toggle LED to indicate UART activity
		mLED_0_Toggle();
	}

	// We don't care about TX interrupt
	if ( mU2TXGetIntFlag() )
	{
		mU2TXClearIntFlag();
	}
}


/** Other Functions ********************************************/

void initInterruptController(void)
{
	// init Timer3 mode and period (PR3) 
	OpenTimer3( T3_ON | T3_PS_1_2 | T3_SOURCE_INT, 0x7A11); // produces ~1000 ms period
															// sampling frequency = ~1 kHz
	mT3SetIntPriority( 7); 	// set Timer3 Interrupt Priority
	mT3ClearIntFlag(); 		// clear interrupt flag
	mT3IntEnable( 1);		// enable timer3 interrupts
}

void initUART2(int pbClk)
{
	 // define setup Configuration 1 for OpenUARTx
		// Module Enable 
		// Work in IDLE mode 
		// Communication through usual pins 
		// Disable wake-up 
		// Loop back disabled 
		// Input to Capture module from ICx pin 
		// no parity 8 bit 
		// 1 stop bit 
		// IRDA encoder and decoder disabled 
		// CTS and RTS pins are disabled 
		// UxRX idle state is '1' 
		// 16x baud clock - normal speed
	#define config1 	UART_EN | UART_IDLE_CON | UART_RX_TX | UART_DIS_WAKE | UART_DIS_LOOPBACK | UART_DIS_ABAUD | UART_NO_PAR_8BIT | UART_1STOPBIT | UART_IRDA_DIS | UART_DIS_BCLK_CTS_RTS| UART_NORMAL_RX | UART_BRGH_SIXTEEN
	 
	 // define setup Configuration 2 for OpenUARTx
		// IrDA encoded UxTX idle state is '0'
		// Enable UxRX pin
		// Enable UxTX pin
		// Interrupt on transfer of every character to TSR 
		// Interrupt on every char received
		// Disable 9-bit address detect
		// Rx Buffer Over run status bit clear
	 #define config2		UART_TX_PIN_LOW | UART_RX_ENABLE | UART_TX_ENABLE | UART_INT_TX | UART_INT_RX_CHAR | UART_ADR_DETECT_DIS | UART_RX_OVERRUN_CLEAR	
 
	// Open UART2 with config1 and config2
	OpenUART2( config1, config2, pbClk/16/DESIRED_BAUDRATE-1);	// calculate actual BAUD generate value.
 		
	// Configure UART2 RX Interrupt with priority 2
	ConfigIntUART2(UART_INT_PR2 | UART_RX_INT_EN);
}


void sendDataRS232()
{
	int i;
	char RS232_Out_Buffer[32]; // max characters per line (line feed and carriage return count)
	
	sprintf(RS232_Out_Buffer,"\n\rSTART\n\r"); //print START. MATLAB uses this as a start delimiter
	putsUART2(RS232_Out_Buffer);
	
	sprintf(RS232_Out_Buffer,"ROWS=%d\n\r", N); //print the number of rows, so matlab can make the correct sized buffer
	putsUART2(RS232_Out_Buffer);
	
	for(i = 0; i < N; i++)
	{
		// output: frequency vector, the calculated fft, raw samples
		sprintf(RS232_Out_Buffer,"%d %d %d %d %d %d %d %d %d\n\r",	dout[i].re, dout[i].im, calcBuffer[i].re, calcBuffer[i].im, freqVector[i], singleSidedFFT[i], FFTconjugate[i].re, FFTconjugate[i].im, FFTPhase[i]); 
															        //IFFT real   IFFT imag   Samples real      Samples imag      FFT freq vect  FFT magn           FFT conj real       FFT conj imag    FFT phas
		putsUART2(RS232_Out_Buffer);
	}
	
	sprintf(RS232_Out_Buffer,"END\n\r"); //output end so matlab knows we're done
	putsUART2(RS232_Out_Buffer);
	
}


void computeFFT()
{
	// when using 256 samples, we measured this function to take about 500 microseconds
	// (not including the time to send rs232 data)
	int i;
	int sum = 0;
	mT3IntEnable(0); //turns off interrupt while computing FFT
	LOOP_TIME_PIN = TRUE;

	for (i=0; i<N; i++)
	{
		if (i<sampleIndex)
		{
			calcBuffer[i+(N-sampleIndex)].re = N * sampleBuffer[i].re;	// old chunk
			calcBuffer[i+(N-sampleIndex)].im = N * sampleBuffer[i].im;	// old chunk
		}	
		else // i >= sampleIndex
		{
			calcBuffer[i-sampleIndex].re = N * sampleBuffer[i].re;		// new chunk
			calcBuffer[i-sampleIndex].im = N * sampleBuffer[i].im;		// new chunk
		}	
	}	
	// load complex input data into din
	mips_fft32(dout, calcBuffer, fftc, scratch, log2N);	
	// compute single sided fft
	for(i = 0; i < N/2; i++)
	{
		singleSidedFFT[i] = sqrt((dout[i].re*dout[i].re) + (dout[i].im*dout[i].im));
	}
	for(i = 0; i < N/2; i++)
	{
		sum = sum + singleSidedFFT[i];
	}
	for(i = 0; i < N/2; i++)
	{
		if(singleSidedFFT[i] > sum/(N/2))
		{
			FFTPhase[i] = atan2(dout[i].im,dout[i].re);
		}
		else
		{
			FFTPhase[i] = 0;
		}
	}

	LOOP_TIME_PIN = FALSE;
	computeFFTflag = FALSE;
	// do something with dout
	mT3IntEnable(1); //turn interrupt back on
}

void computeIFFT()
{
	// computes IFFT
	int i;
	mT3IntEnable(0); //turns off interrupt while computing FFT
	LOOP_TIME_PIN = TRUE;
	for (i=0; i<N; i++)
	{
		FFTconjugate[i].re = N * dout[i].re;
		FFTconjugate[i].im = N * (0 - dout[i].im);
	}	

	// load complex input data into din
	mips_fft32(dout, FFTconjugate, fftc, scratch, log2N);
	// compute single sided fft

	LOOP_TIME_PIN = FALSE;
	computeFFTflag = FALSE;
	mT3IntEnable(1); //turn interrupt back on
}

void initPWM(void)
{
	// init OC1 module, on pin D0
	OpenOC1( OC_ON | OC_TIMER2_SRC | OC_PWM_FAULT_PIN_DISABLE, 0, 0);
	
	// init Timer2 mode and period PR2
	// if set for 20kHz, PR=3999=0x0F9F, Pre = 1
	OpenTimer2( T2_ON | T2_PS_1_1 | T2_SOURCE_INT, 0x0F9F); // every 1/20000 seconds

	SetDCOC1PWM(20833);
}

void initPWMInterrupt(void)
{
	OpenTimer4( T4_ON | T4_PS_1_1 | T4_SOURCE_INT, 0x1E84); // every 1/freq/PTS_PER_DUTY seconds
	mT4SetIntPriority( 6);
	mT4ClearIntFlag();
	mT4IntEnable( 1);
}

void initADC(void)
{
	// configure and enable the ADC
	CloseADC10();	// ensure the ADC is off before setting the configuration

	// define setup parameters for OpenADC10
	// Turn module on | output in integer | trigger mode auto | enable  autosample
	#define PARAM1  ADC_MODULE_ON | ADC_FORMAT_INTG16 | ADC_CLK_AUTO | ADC_AUTO_SAMPLING_ON
	// ADC ref external    | disable offset test    | enable scan mode | perform 2 samples | use one buffer | use MUXA mode
	#define PARAM2  ADC_VREF_AVDD_AVSS | ADC_OFFSET_CAL_DISABLE | ADC_SCAN_ON | ADC_SAMPLES_PER_INT_2 | ADC_ALT_BUF_OFF | ADC_ALT_INPUT_OFF
	// Use ADC internal clock | set sample time
	#define PARAM3  ADC_CONV_CLK_INTERNAL_RC | ADC_SAMPLE_TIME_15
	// Set AN4
	#define PARAM4	ENABLE_AN4_ANA
	// Do not assign channels to scan
	#define PARAM5	SKIP_SCAN_AN0 | SKIP_SCAN_AN1 | SKIP_SCAN_AN2 | SKIP_SCAN_AN3 | SKIP_SCAN_AN5 | SKIP_SCAN_AN6 | SKIP_SCAN_AN7 | SKIP_SCAN_AN8 | SKIP_SCAN_AN9 | SKIP_SCAN_AN10 | SKIP_SCAN_AN11 | SKIP_SCAN_AN12 | SKIP_SCAN_AN13 | SKIP_SCAN_AN14 | SKIP_SCAN_AN15

	// use ground as neg ref for A | use AN4 (B4) for input A
	// configure to sample AN4
	SetChanADC10( ADC_CH0_NEG_SAMPLEA_NVREF); // configure to sample AN4
	OpenADC10( PARAM1, PARAM2, PARAM3, PARAM4, PARAM5 ); // configure ADC using parameter define above

	EnableADC10(); // Enable the ADC

	while ( ! mAD1GetIntFlag() ) { } // wait for the first conversion to complete so there will be valid data in ADC result registers	
}

void initSinPWM(void)
{
	int i;
	float dt, amp, offset;
	float amp2, offset2;
	float fundFreq = 10;		//10 is GCD of 30 and 40, freq and freq2

	// Make sine wave PWM vector
	dt = 1/fundFreq/PTS_PER_DUTY;
	amp = 0.817 * mVpp + 40.833;
	offset = amp;
	amp2 = 0.817 * mVpp2 + 40.833;
	offset2 = amp2;

	for (i=0; i<PTS_PER_DUTY; i++) {
		duty_cycles[i] = amp*sin(2*PI*freq*(i*dt)+phase)+offset + amp*sin(2*PI*freq2*(i*dt)+phase)+offset;
		if (duty_cycles[i] < 1) {
			duty_cycles[i] = 1;
		}
	}
}

